/* SPIM S20 MIPS simulator.
   Description of a SPIM S20 instruction.
   (Layout does not correspond to MIPS machine.)
   Copyright (C) 1990 by James Larus (larus@cs.wisc.edu).

   SPIM is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 1, or (at your option) any
   later version.

   SPIM is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with GNU CC; see the file COPYING.  If not, write to James R.
   Larus, Computer Sciences Department, University of Wisconsin--Madison,
   1210 West Dayton Street, Madison, WI 53706, USA or to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */


/* $Header: /var/home/larus/Courses/536/SPIM/RCS/inst.h,v 3.0 90/08/24 15:10:29 larus Exp Locker: larus $
*/



/* Describes an expression that produce a value for an instruction's
   immediate field.  Immediates have the form: label +/- offset. */

typedef struct immexpr
{
  long offset;			/* Offset from symbol */
  struct lab *symbol;		/* Symbolic label */
  short bits;			/* > 0 => 31..16, < 0 => 15..0 */
  short pc_relative;		/* Non-zero => offset from label in code */
} imm_expr;


/* Describes an expression that produce an address for an instruction.
   Address have the form: label +/- offset (register). */

typedef struct addrexpr
{
  unsigned char reg_no;		/* Register number */
  imm_expr *imm;		/* The immediate part */
} addr_expr;



/* Choose bland names for fields and #define the real names below.
   This is better than a union since it doesn't require an additional
   level of syntactic indirection in references. */

typedef struct inst_s
{
  short opcode;
  unsigned char f1;
  unsigned char f2;
  unsigned char f3;
  unsigned char f4;
  long f5;
  short f5x;
  imm_expr *f6;
} instruction;


#define OPCODE(INST)	(INST)->opcode

#define RS(INST)	(INST)->f1
#define FS(INST)	(INST)->f1
#define BASE(INST)	(INST)->f1

#define RT(INST)	(INST)->f2
#define FT(INST)	(INST)->f2

#define RD(INST)	(INST)->f3
#define FD(INST)	(INST)->f3

#define SHAMT(INST)	(INST)->f4
#define FORMAT(INST)	(INST)->f4

#define IMM(INST)	(INST)->f5x
#define IOFFSET(INST)	(INST)->f5x
#define COND(INST)	(INST)->f5x

#define TARGET(INST)	(INST)->f5

#define EXPR(INST)	(INST)->f6


#define COND_UN		0x1
#define COND_EQ		0x2
#define COND_LT		0x4
#define COND_IN		0x8



/* Raise an exception! */

#define RAISE_EXCEPTION(CAUSE, MISC)					\
	{								\
	  if (((CAUSE)<= LAST_REAL_EXCEPT) || (Status_Reg & 0x1))	\
	    {								\
	      Cause = (CAUSE) << 2;					\
	      exception_occurred = 1;					\
	      EPC = PC;							\
	      Status_Reg = Status_Reg & 0xffffffc0 | (Status_Reg & 0xf) << 2; \
	      MISC;							\
	    }								\
	}								\


/* Recognized exceptions (see Ch. 5): */

#define INT_EXCPT 0
#define ADDRL_EXCPT 4
#define ADDRS_EXCPT 5
#define IBUS_EXCPT 6
#define DBUS_EXCPT 7
#define SYSCALL_EXCPT 8
#define BKPT_EXCPT 9
#define RI_EXCPT 10
#define OVF_EXCPT 12


/* Floating point exceptions (Ch. 8): */

#define INEXACT_EXCEPT 13
#define INVALID_EXCEPT 14
#define DIV0_EXCPET 15
#define FOVF_EXCEPT 16
#define FUNF_EXCEPT 17

#define LAST_REAL_EXCEPT FUNF_EXCEPT


/* External interrupts: */

#define CLOCK_EXCPT 0x100       /* IP[0] */
#define IO_EXCPT 0x200          /* IP[1] */
